"""
Project: BlueMath_tk
Sub-Module: distributions
Author: GeoOcean Research Group, Universidad de Cantabria
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: Under development (Working)
"""

# Import essential functions/classes to be available at the package level.
from .gev import GEV
from .gpd import GPD
from .nonstat_gev import NonStatGEV

# Optionally, define the module's `__all__` variable to control what gets imported when using `from module import *`.
__all__ = ["GEV", "GPD", "NonStatGEV"]
